/*
 *   Copyright (c) 2005 - 2011 by Thanos Kyritsis
 *
 *   This file is part of ffpocket
 *
 *   FFpocket is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2 of the License.
 *
 *   FFpocket is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with FFpocket; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "ffdecode.h"


AVStream *ff_add_audio_stream(AVFormatContext *oc, enum CodecID codec_id, AVCodecContext *ccx) {
	AVCodecContext *c;
	AVStream *st;
	int retflag = TRUE;

	st = av_new_stream(oc, 1);
	if (!st) {
		fprintf(stderr, "Could not alloc stream\n");
		retflag = FALSE;
	}
	
	if (retflag != FALSE) {
		c = st->codec;
		c->codec_id = codec_id;
		c->codec_type = AVMEDIA_TYPE_AUDIO;

		/* put sample parameters */
		//c->bit_rate = 64000;
		c->sample_rate = ccx->sample_rate;
		c->channels = ccx->channels;
		c->sample_fmt = ccx->sample_fmt;

		// some formats want stream headers to be separate
		if(oc->oformat->flags & AVFMT_GLOBALHEADER)
			c->flags |= CODEC_FLAG_GLOBAL_HEADER;
	}
	
	if(retflag != FALSE)	
		return st;
	else
		return NULL;
}

int ff_decode_audio(const char* outfilename, const char *filename, 
					 const char *timepos_start, const char *timepos_stop, const int rangedecoding,
					 const char *cachefilename) {
	AVCodec *codec=NULL, *vcodec=NULL, *codec_enc=NULL;
	AVPacket pkt;
	AVCodecContext *ccx=NULL;
	AVFormatContext *fcx=NULL, *ofcx=NULL;
	AVOutputFormat *fmt=NULL;
	AVStream *audio_st=NULL;
	int got_audio, len, err=0, audio_index=-1, i=0, counter=0, j=0, bytecounter=0, channels=-1, pack_duration=0, packsize=0, codec_is_open=-1, video_index=-1, codec_enc_is_open=-1;
	char sample;
	int step = sizeof(sample)*2;
	char *maxsample=NULL, *minsample=NULL;
	int charsize = sizeof(char);
	float ratewindow=-1;
	unsigned int offset=0, maxbyte=0, sampledcounter=1;
	long long int seek_time_start=0, seek_time_stop=0, pack_pts=0;
	FILE *cachefile=NULL;
	char *outbuf;
	unsigned char *packptr;
	int ret = TRUE, nobrk = TRUE;
	int endianess = isLittleEndian();

	av_register_all();

	// Open the input/output files
	err = av_open_input_file(&fcx, filename, NULL, 0, NULL);
	if(err<0){
		fprintf(stderr, "Can't open file: %s\n", filename);
		ret = FALSE;
	}
	outbuf = av_malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
	if(outbuf==NULL) {
		fprintf(stderr, "Cannot allocate memory for outbuf.\n");
		ret = FALSE;
	}
	
	if(cachefilename) {
		cachefile = fopen(cachefilename, "wb");
		if (!cachefile) {
			fprintf(stderr, "Could not open '%s'\n", cachefilename);
			ret = FALSE;
		}
	}

	if (ret != FALSE) {	
		// Find the stream info.
		av_find_stream_info(fcx);

		/* Find the first supported codec in the audio streams */
		for(i=0; i<fcx->nb_streams; i++) {
			ccx=fcx->streams[i]->codec;
			if(ccx->codec_type==AVMEDIA_TYPE_AUDIO) {
				/* Found an audio stream, check if codec is supported */
				codec = avcodec_find_decoder(ccx->codec_id);
				if(codec){
					/* codec is supported, proceed */
					audio_index = i;
					break;
				}
			}
		}
		/* Find the first supported video codec in the stream
		 * This is only meant for seeking. When we have muxed
		 * audio and video stream we have to seek by the video
		 */
		for(i=0; i<fcx->nb_streams; i++) {
			if(fcx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO) {
				/* Found a video stream, check if codec is supported */
				vcodec = avcodec_find_decoder(fcx->streams[i]->codec->codec_id);
				if(vcodec) {
					/* codec is supported, proceed */
					video_index = i;
					break;
				}
			}
		}
		/* If we didn't find any supported video stream, it might mean
		 * we are dealing with an audio only stream, so we can use the 
		 * audio_index for seeking */
		if (video_index < 0) video_index = audio_index;
	}

	/* Find codec id */
	if(audio_index < 0) {
	 	fprintf(stderr, "Audio stream with supported codec not found.\n");
	 	ret = FALSE;
	}
	else {
		/* open it */
		if ((codec_is_open = avcodec_open(ccx, codec)) < 0) {
			fprintf(stderr, "could not open codec\n");
			ret = FALSE;
		}
		else {
			// See how many channels is the input file
			if(cachefilename) {
				channels = ccx->channels;
				ratewindow = (float)ccx->sample_rate / RESOLUTION;
				maxbyte = (sampledcounter) * channels * step * ratewindow;
				maxsample = malloc(channels);
				if(maxsample==NULL) {
					fprintf(stderr, "Could not allocate memory for maxsample.\n");
					ret = FALSE;
				}
				minsample = malloc(channels);
				if(minsample==NULL) {
					fprintf(stderr, "Could not allocate memory for minsample.\n");
					ret = FALSE;
				}
		
				if (ret != FALSE) {
					// Initialize these arrays
					for (j=1;j<=channels;j++) {
						minsample[j-1] = 127;
						maxsample[j-1] = -128;
					}
					fprintf(stdout, "Audio caching...\n\n");
				}
			}

			if (rangedecoding) {
				seek_time_start = parse_date(timepos_start,1);
				seek_time_stop = parse_date(timepos_stop,1);
				if (fcx->start_time != AV_NOPTS_VALUE) {
					seek_time_start += fcx->start_time;
					seek_time_stop += fcx->start_time;
				}
				/* Check that timepos_start is smaller than timepos_stop */
				if (seek_time_start < seek_time_stop) {
					/* Check that timepos is inside duration */
					if (seek_time_start <= fcx->duration && seek_time_stop <= fcx->duration) {
						/* All ok */
						/*  Seek to the nearest keyframe before the seek_time we asked */
						/* Notice: always seek to video stream because if it is an avi 
						 * we need to seek by the video stream. not the audio */
						av_seek_frame(fcx, video_index, av_rescale_q(seek_time_start, AV_TIME_BASE_Q, fcx->streams[video_index]->time_base), AVSEEK_FLAG_BACKWARD);
					}
					else {
						fprintf(stderr, "Seek start/stop time cannot be greater than input file's duration.\nExiting...\n");
						ret = FALSE;
					}
				}
				else {
					fprintf(stderr, "Seek start time is not smaller than seek stop time.\nExiting ...\n");
					ret = FALSE;
				}
			}
		}
	}

	/* Create wav headers */
	if (!cachefilename && ret != FALSE) {
		// allocate the output media context
		ofcx = avformat_alloc_context();

		if (!ofcx) {
			fprintf(stderr, "Memory error\n");
			ret = FALSE;
		}

		// use wav as the output format of the file
		fmt = av_guess_format(NULL, outfilename, NULL);
		if (!fmt) {
			fprintf(stderr, "Could not deduce output format from file extension: using WAV.\n");
			fmt = av_guess_format("wav", NULL, NULL);
		}
		if (!fmt) {
			fprintf(stderr, "Could not find suitable output format\n");
			ret = FALSE;
		}

		if(ret != FALSE) {
			ofcx->oformat = fmt;
			snprintf(ofcx->filename, sizeof(ofcx->filename), "%s", outfilename);

			// add the audio stream using the default format codec and initialize the codec
			if (fmt->audio_codec != CODEC_ID_NONE) {
				// uncomment if you want downmix to stereo
				/*if (ccx->channels > 2) {
					ccx->channels = 2;
					ccx->request_channels = 2;
				}*/
				audio_st = ff_add_audio_stream(ofcx, fmt->audio_codec, ccx);
			}

			// set the output parameters (must be done even if no parameters)
			if (av_set_parameters(ofcx, NULL) < 0) {
				fprintf(stderr, "Invalid output format parameters\n");
				ret = FALSE;
			}
			else {
				dump_format(ofcx, 0, outfilename, 1);

				codec_enc = avcodec_find_encoder(audio_st->codec->codec_id);
				if (!codec_enc) {
					fprintf(stderr, "encoder codec not found\n");
					ret = FALSE;
				}
				else {
					if ((codec_enc_is_open = avcodec_open(audio_st->codec, codec_enc)) < 0) {
						fprintf(stderr, "could not open encoder codec\n");
						ret = FALSE;
					}
					else {
						if (!(fmt->flags & AVFMT_NOFILE)) {
							if (url_fopen(&ofcx->pb, outfilename, URL_WRONLY) < 0) {
								fprintf(stderr, "Could not open '%s'\n", outfilename);
								ret = FALSE;
							}
							else {
								av_write_header(ofcx);
								fprintf(stdout, "Audio decoding...\n\n");
							}
						}
					}
				}
			}
		}
	}

	while(ret != FALSE && nobrk != FALSE) {
		// Read a frame/packet
		if(av_read_frame(fcx, &pkt) < 0 ) break;
		packsize = pkt.size;
		packptr = pkt.data;
		// Make sure this packet belongs to the stream we want
		if(pkt.stream_index==audio_index) {
			while (packsize > 0 && nobrk != FALSE) {
				// Rescale the times
				pack_pts = av_rescale_q(pkt.pts, fcx->streams[audio_index]->time_base, AV_TIME_BASE_Q);
				pack_duration = av_rescale_q(pkt.duration, fcx->streams[audio_index]->time_base, AV_TIME_BASE_Q);

				// uncomment if you want downmix to stereo
				/*if (ccx->channels > 2) {
					ccx->channels = 2;
					ccx->request_channels = 2;
				}*/
				// Decode the paket
				got_audio = AVCODEC_MAX_AUDIO_FRAME_SIZE;
				len = avcodec_decode_audio3(ccx, (short *)outbuf, &got_audio, &pkt);

				if (len < 0) {
					fprintf(stderr, "Error while decoding\n");
					break;
				}

				packsize -= len;
				packptr += len;

				if ( got_audio > 0 ) {
					// If we got the time exactly, or we are already past the seek time,
					// this is the frame we want
					if (rangedecoding && pack_pts >= seek_time_start) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_enc_out(ofcx, audio_st, (short *)outbuf, got_audio);

						// if the next frame gets past our stop time, we want to stop decoding
						if ( pack_pts + pack_duration > seek_time_stop ) {
							av_free_packet(&pkt);
							nobrk = FALSE;
							break;
						}
					}
					// If the next frame will be past our start seek time, this is the frame we want
					else if (rangedecoding && pack_pts + pack_duration > seek_time_start) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_enc_out(ofcx, audio_st, (short *)outbuf, got_audio);

						// if the next frame gets past our stop time, we want to stop decoding
						if ( pack_pts + pack_duration > seek_time_stop ) {
							av_free_packet(&pkt);
							nobrk = FALSE;
							break;
						}
					}
					else if (!rangedecoding) {
						// if a frame has been decoded, output it
						if(!cachefilename) ff_audio_enc_out(ofcx, audio_st, (short *)outbuf, got_audio);
					}

					// Output cache
					if(cachefilename) {
						// bytecounter counts for each 'chunk' avcodec gives us, not the whole stream
						// bytecounter+offset is our current position in the whole stream
						bytecounter=0;
						// got_audio is the length of the 'chunk'
						while (bytecounter < got_audio) {
							// we check if the byte of the next sample is beyond the byte of the downsampling window
							if (bytecounter+offset + step*channels >= maxbyte) {
								for (j = 1; j <= channels ; j++) {
									// actually sample is not the whole sample, but the data for one channel each time
									sample = (char)outbuf[bytecounter+endianess];
									// min max averaging: only keep the highest and the lowest sample value
									if (maxsample[j-1] < sample) maxsample[j-1] = sample;
									if (minsample[j-1] > sample) minsample[j-1] = sample;
									fwrite(&maxsample[j-1], charsize, 1, cachefile);
									fwrite(&minsample[j-1], charsize, 1, cachefile);
									// reset min and max
									minsample[j-1] = 127;
									maxsample[j-1] = -128;
									// move 'chunk' byte pointer to current sample, next channel
									bytecounter += step;
								}
								// increase the counter showing how many samples we've averaged
								sampledcounter++;
								// move downsampling window pointer to the next window position
								maxbyte = (sampledcounter) * channels * step * ratewindow;
							}
							else {
								for (j = 1; j <= channels; j++) {
									sample = (char)outbuf[bytecounter+endianess];
									// min max averaging
									// if you want only downsampling, comment min max
									if (maxsample[j-1] < sample) maxsample[j-1] = sample;
									if (minsample[j-1] > sample) minsample[j-1] = sample;
									// move 'chunk' byte pointer to current sample, but next channel
									bytecounter += step;
								}
							}
						}
					}
					// Display the timestamp as progress (eye candy)
					ff_display_decoding_progress(pack_pts, counter);
					counter++;
					// add to the offset our current position
					offset += bytecounter;
				}
			}
		}
		av_free_packet(&pkt);
	}

	/* Clean up */
	fprintf(stdout, "\n");
	
	if (!cachefilename) {
		if (ret != FALSE) {
			/* write the trailer, if any */
			av_write_trailer(ofcx);
		}

		/* close each codec */
		if (codec_enc_is_open >= 0) avcodec_close(audio_st->codec);

		if (ofcx != NULL) {
			/* free the streams */
			for(i = 0; i < ofcx->nb_streams; i++) {
				av_freep(&ofcx->streams[i]->codec);
				av_freep(&ofcx->streams[i]);
			}

			/* close the output file */
			if (!(fmt->flags & AVFMT_NOFILE) && ofcx->pb != NULL) {
				url_fclose(ofcx->pb);
			}

			/* free the stream */
			av_free(ofcx);
		}
	}

	if (cachefilename) {
		if(maxsample != NULL) free(maxsample);
		if(minsample != NULL) free(minsample);
		if(cachefile != NULL) fclose(cachefile);
	}
	
	if (codec_is_open >= 0) avcodec_close(ccx);
	if (outbuf != NULL)     av_free(outbuf);
	if (fcx != NULL)        av_close_input_file(fcx);

	return ret;
}

void ff_audio_enc_out(AVFormatContext *ofcx, AVStream *audio_st, const short *samples, int buf_size) {
	AVPacket pkt;
	uint8_t *outbuf;
	int ret = TRUE;

	av_init_packet(&pkt);
	
	outbuf = av_malloc(buf_size);
	if(outbuf==NULL) {
		fprintf(stderr, "Cannot allocate memory for output encoded buffer.\n");
		ret = FALSE;
	}

	if (ret != FALSE) {
		pkt.size = avcodec_encode_audio(audio_st->codec, outbuf, buf_size, samples);

		pkt.stream_index= audio_st->index;
		pkt.data = outbuf;
		if(audio_st->codec->coded_frame && audio_st->codec->coded_frame->pts != AV_NOPTS_VALUE)
			pkt.pts = av_rescale_q(audio_st->codec->coded_frame->pts, audio_st->codec->time_base, audio_st->time_base);
		pkt.flags |= AV_PKT_FLAG_KEY;
							
		/* write the compressed frame in the media file */
		if (av_write_frame(ofcx, &pkt) != 0) {
		  fprintf(stderr, "Error while writing audio frame\n");
		}

		if (outbuf != NULL)
			av_free(outbuf);

		av_free_packet(&pkt);
	}
}

